# -*- coding: utf-8 -*-
'''
NetworkDesign/mod_models/ModBaseModel

@author: Bela Genge
'''

from mod_network import ModNetwork
from mod_flow import ModFlow
from mod_link import ModLink

class ModBaseModel(ModNetwork):
    '''
    '''

    def __init__(self, name):
        '''
        Constructor
        '''
        super(ModBaseModel, self).__init__(name)

        self._gmap = {}
        self._xmap = {}
        self._wmap = {}
        self._tmap = {}
        self._ymap = {}
        self._fmap = {}
        self._cctr = 0 # Constraint counter
        self.__lpfname = './NetworkDesign.lp'
        self.__vctr = 0 # Var counter - used to place newline after a number of variables


    def generateLPFile(self, fname=None):
        '''
        Function used to generate the LP model file from the network model
        '''

        # Create/open the file for writing
        if fname is None:
            lpf = open(self.__lpfname, "w")
        else:
            lpf = open(fname, "w")

        # Write header
        lpf.write("\\ENCODING=ISO-8859-1\n\\Problem name: CPLEX-API\n\n")

        # Generate t_jlbi variable indexes
        if len(self._tmap) <= 0:
            self._gen_varMapping()

        # Write objective
        self._writeObjective(lpf)

        # Write constraints
        self._writeConstraints(lpf)

        # Write bounds and binary variables
        self._writeCBoundsBinaries(lpf)

        # End LP file
        lpf.write("End\n")

        self._printTmap()

        # Close file
        lpf.close()


    def getSolutionMapping(self, vars):

        # Get first mapping for the solution variables
        map_x = {}
        map_w = {}
        map_t = {}
        map_y = []
        map_f = []

        for (v,val) in vars:

            xm = self._get_x_varMapping(v)
            wm = self._get_w_varMapping(v)
            tm = self._get_t_varMapping(v)
            ym = self._get_y_varMapping(v)
            fm = self._get_f_varMapping(v)

            if xm is not None:
                (i,j) = xm
                map_x[i] = j
            if wm is not None:
                (j,i) = wm
                map_w[i] = j
            if tm is not None:
                (j,l,i) = tm
                if i not in map_t:
                    map_t[i] = [(j,l)]
                else:
                    map_t[i].append((j,l))
            if ym is not None:
                map_y.append(ym)
            if fm is not None:
                map_f.append((fm,val))

        # Next, order the nodes in the t-mapping so that we get the path
        map1 = {}
        for (i,map) in map_t.iteritems():
            map1[i] = []

            # Start from Access node and step through the list until we reach Egress node
            n = map_x[i]  
            ne = map_w[i]
            while n != ne:
                for (j,l) in map:
                    if n == j:
                        map1[i].append((j,l))
                        n = l
            #if ne not in fmap1[i]:
            if len(map1[i]) <= 0:
                map1[i].append((ne,ne)) # For the last node we do not have a specific link

        # For the flow end-points we need the actual nodes on which they are installed
        map_nx = {}
        map_nw = {}
        for (i,j) in map_x.iteritems():
            map_nx[i] = (j,self._flows[i].getProp(ModFlow.FACC))
        for (i,j) in map_w.iteritems():
            map_nw[i] = (j,self._flows[i].getProp(ModFlow.FEGG))        

        return (map1, map_nx, map_nw, map_y, map_f)


    def _printTmap(self):

        '''self._gmap = {}
        self._xmap = {}
        self._wmap = {}
        self._tmap = {}
        self._ymap = {}
        self._fmap = {}
        '''

        print "*** g-Variable mapping: "
        for ctr in range(1,len(self._cnodes)+1):
            for (j,idx) in self._gmap.iteritems():
                if idx == ctr:
                    print "x", ctr, ": ", j
                    break

        st = len(self._cnodes)
        print "*** x-Variable mapping: "
        for ctr in range(1,len(self._flows)*len(self._cnodes)+1):
            for (i,j),idx in self._xmap.iteritems():
                if idx == st+ctr:
                    print "x", st+ctr, ": ", i, j
                    break

        st += len(self._cnodes)*len(self._flows)
        print "*** w-Variable mapping: "
        for ctr in range(1,len(self._flows)*len(self._cnodes)+1):
            for (j,i),idx in self._wmap.iteritems():
                if idx == st+ctr:
                    print "x", st+ctr, ": ", j, i
                    break

        print "*** t-Variable mapping: "
        st += len(self._flows)*len(self._cnodes)
        for ctr in range(1,len(self._cnodes)*len(self._cnodes)*len(self._flows)+1):
            for (j,l,i),idx in self._tmap.iteritems():
                if idx == st+ctr:
                    print "x", st+ctr, ": ", j, l, i
                    break

        st += len(self._cnodes)*len(self._cnodes)*len(self._flows)
        print "*** y-Variable mapping: "
        for ctr in range(1,len(self._cnodes)*self._maxsz+1):
            for (j,s),idx in self._ymap.iteritems():
                if idx == st+ctr:
                    print "x", st+ctr, ": ", j, s
                    break

        st += len(self._cnodes)*self._maxsz
        print "*** f-Variable mapping: "
        for ctr in range(1,len(self._cnodes)*len(self._cnodes)+1):
            for (j,l),idx in self._fmap.iteritems():
                if idx == st+ctr:
                    print "x", st+ctr, ": ", j, l
                    break


    def _get_t_varMapping(self, var):

        # This should be done with a MATH formula, not like this!!! - the problem is that stuff is in dictionaries, the order is not the one that it was used for adding them.
        for j,jobj in self._cnodes.iteritems():
            for l,lobj in self._cnodes.iteritems():
                for i,iobj in self._flows.iteritems():
                    if self._tmap[(j,l,i)] == var:
                        return (j,l,i)
        return None


    def _get_x_varMapping(self, var):

        # This should be done with a MATH formula, not like this!!!
        for i,iobj in self._flows.iteritems():
            for j,jobj in self._cnodes.iteritems():
                if self._xmap[(i,j)] == var:
                    return (i,j)

        return None


    def _get_w_varMapping(self, var):
        
        for j,jobj in self._cnodes.iteritems():
            for i,iobj in self._flows.iteritems():
                if self._wmap[(j,i)] == var:
                    return (j,i)

        return None


    def _get_y_varMapping(self, var):

        for j,jobj in self._cnodes.iteritems():
            for s in range(1, self._maxsz+1):
                if self._ymap[(j,s)] == var:
                    return (j,s)

        return None


    def _get_f_varMapping(self, var):

        # This should be done with a MATH formula, not like this!!! - the problem is that stuff is in dictionaries, the order is not the one that it was used for adding them.
        for j,jobj in self._cnodes.iteritems():
            for l,lobj in self._cnodes.iteritems():
                if self._fmap[(j,l)] == var:
                    return (j,l)
        return None


    def _gen_varMapping(self):
        # Generate the mapping between network names and variable names
        self._gmap = {}
        self._xmap = {}
        self._wmap = {}
        self._tmap = {}
        self._ymap = {}
        self._fmap = {}

        vidx = 1

        # Generate mapping for g - nodes of type CoS
        for j,jobj in self._cnodes.iteritems():
            self._gmap[j] = vidx
            vidx += 1

        # Generate mapping for x
        for i,iobj in self._flows.iteritems():
            for j,jobj in self._cnodes.iteritems():
                self._xmap[(i,j)] = vidx
                vidx += 1

        # Generate mapping for w
        for j,jobj in self._cnodes.iteritems():
            for i,iobj in self._flows.iteritems():
                self._wmap[(j,i)] = vidx
                vidx += 1

        # Generate mapping for t
        for j,jobj in self._cnodes.iteritems():
            for l,lobj in self._cnodes.iteritems():
                for i,iobj in self._flows.iteritems():
                    self._tmap[(j,l,i)] = vidx
                    vidx += 1

        # Generate mapping for y
        for j,jobj in self._cnodes.iteritems():
            for s in range(1, self._maxsz+1):
                self._ymap[(j,s)] = vidx
                vidx += 1

        # Generate mapping for f
        for j,jobj in self._cnodes.iteritems():
            for l,lobj in self._cnodes.iteritems():
                self._fmap[(j,l)] = vidx
                vidx += 1


    def _writeObjective(self, lpf):
        '''
        Write the objective function.
        '''
        pass


    def _writeConstraints(self, lpf):
        '''
        Write constraints.
        '''

        self._cctr = 0
        lpf.write("Subject To\n")

        # Write flow conservation constraints
        self._writeCFlowConservation(lpf)

        # Write link capacity constraints
        # TODO: we should add concentrator capacity constraints as well
        self._writeCLinkCapacity(lpf, True)
        self._writeCLinkCapacity(lpf, False)        


    def _writeCFlowConservation(self, lpf):
        '''
        Write flow conservation constraints: x_ij - w_ji - sum_l(t_jli - t_lji) = 0
        that is: - sum_l(t_jli - t_lji) = w_ji - x_ij
        '''

        for j,jobj in self._cnodes.iteritems():
            for i,iobj in self._flows.iteritems():

                # Prepare list of variables for ordering - so it is easier to read
                # Make sure we do not lose the sign
                tvars = []
                tvarsign = {}
                for l,lobj in self._cnodes.iteritems():

                    # In the case of j = l we skip and go to the next variable
                    if j == l:
                        continue

                    tvars.append(self._tmap[(j,l,i)])
                    tvarsign[self._tmap[(j,l,i)]] = -1

                    tvars.append(self._tmap[(l,j,i)])
                    tvarsign[self._tmap[(l,j,i)]] = 1

                # Perform the ordering
                tvars.sort()

                # Write the constraint
                self._cctr += 1
                lpf.write(" c" + str(self._cctr) + ":  ")

                first = True
                for x in tvars:
                    if tvarsign[x] == -1:
                        lpf.write("- ")
                    elif first == False:
                        lpf.write("+ ")
                    lpf.write("x" + str(x) + " ")
                    first = False

                lpf.write("+ x")
                lpf.write(str(self._xmap[(i,j)]))
                lpf.write(" - x")
                lpf.write(str(self._wmap[(j,i)]))
                lpf.write(" = 0\n")


    def _writeCLinkCapacity(self, lpf, nodeJ):
        '''
        Write link capacity constraints: sum_i(t_jli*d_i) - u_jl*b_jl*g_j <= 0
        '''

        for j,jobj in self._cnodes.iteritems():
            for l,lobj in self._cnodes.iteritems():

                if j == l:
                    continue

                # Write the constraint number
                self._cctr += 1
                lpf.write(" c" + str(self._cctr) + ":  ")

                # Write the sum of variables for t_jlbi*d_i
                first = True
                for i,iobj in self._flows.iteritems():
                    if first == True:
                        lpf.write(str(iobj.getProp(ModFlow.DEM)) + " ")
                    else:
                        lpf.write("+ ")
                        lpf.write(str(iobj.getProp(ModFlow.DEM)) + " ")
                    lpf.write("x" + str(self._tmap[(j,l,i)]) + " ")
                    first = False


                ll = self.getLink(j,l)
                if ll != None:
                    if nodeJ == True:
                        lpf.write("- " + str(self._links[ll].getProp(ModLink.LCAP)) + " x" + str(self._gmap[j]))
                    else:
                        lpf.write("- " + str(self._links[ll].getProp(ModLink.LCAP)) + " x" + str(self._gmap[l]))

                # Write the inequality
                lpf.write("<= 0")

                lpf.write("\n")


    def _writeCBoundsBinaries(self, lpf):
        '''
        Write bounds and binary variables - all variables actually.
        '''

        varcount = len(self._cnodes) + 2*len(self._cnodes)*len(self._flows) + len(self._cnodes)*len(self._cnodes)*len(self._flows) + len(self._cnodes)*self._maxsz

        lpf.write("Bounds\n")
        # Write bounds for g,x,w,t,y
        for ctr in range(1, varcount + 1):
            lpf.write(" 0 <= x" + str(ctr) + " <= 1\n")

        # Write bounds for f (integer - the maximum possible conduit value)
        for ctr in range(varcount + 1, varcount + len(self._cnodes)*len(self._cnodes) + 1):
            lpf.write(" 0 <= x" + str(ctr) + " <= " + str(self._maxsc+1) + "\n")

        lpf.write("Binaries\n")
        self._resetLPNewLineCtr()
        for ctr in range(1, varcount + 1):
            lpf.write(" x" + str(ctr) + " ")
            self._addLPNewLine(lpf)
        lpf.write("\n")


    def _addLPNewLine(self, lpf):

        # Write max 9 variables per line - AIMMS style
        self.__vctr = self.__vctr + 1
        if self.__vctr > 9:
            lpf.write("\n     ")
            self.__vctr = 0;


    def _resetLPNewLineCtr(self):

        self.__vctr = 0

