# -*- coding: utf-8 -*-
'''
NetworkDesign/mod_models/ModSecDesignModel
A model used to design security-aware NCI networks.

@author: Bela Genge
'''

from mod_models import ModBaseModel
from mod_node import ModNode
from mod_flow import ModFlow
from mod_link import ModLink

class ModSecDesignModel(ModBaseModel):
    '''
    '''

    def __init__(self, name):
        '''
        Constructor
        '''
        super(ModSecDesignModel, self).__init__(name)


    def _writeObjective(self, lpf):
        '''
        Write the objective function: sum_j c_j^S*g_j + sum_jli c_jl^B*t_jl^i*d_i + sum_ji(c_ij^A*x_ij+c_ji^E*w_ji)*d_i+
                                        + sum_js c_js^Z*y_js + sum_jl c_jl^C*f_jl
        '''
        super(ModSecDesignModel, self)._writeObjective(lpf)

        lpf.write("Minimize\n obj:")

        # Write the node cost: sum_j c_j^S*g_j
        self._resetLPNewLineCtr()
        first = True
        for j,jobj in self._cnodes.iteritems():

            # In the case the node's cost is 0, do not add it to the objective
            if jobj.getProp(ModNode.NCOST) == '0':
                continue

            self._addLPNewLine(lpf)

            # Write the cost value
            if first == True:
                lpf.write(jobj.getProp(ModNode.NCOST) + " ")
                first = False
            else:
                lpf.write("+ " + jobj.getProp(ModNode.NCOST) + " ")

            # Write g_j variable name
            lpf.write("x" + str(self._gmap[j]) + " ")

        # Write the bandwidth per link cost: sum_jli (c_jl^B*t_jl^i*d_i)
        for j,jobj in self._cnodes.iteritems():
            for l,lobj in self._cnodes.iteritems():
                for i,iobj in self._flows.iteritems():

                    if j == l:
                        continue

                    ll = self.getLink(j,l)
                    if ll == None:
                        continue

                    self._addLPNewLine(lpf)

                    val = float(self._links[ll].getProp(ModLink.LCOST))*float(iobj.getProp(ModFlow.DEM))
                    lpf.write(" + " + str(val) + " x" + str(self._tmap[(j,l,i)]))

        # Write the access/egress bandwidth cost: sum_ji(c_ij^A*x_ij+c_ji^E*w_ji)*d_i
        for j,jobj in self._cnodes.iteritems():
            for i,iobj in self._flows.iteritems():

                # For each flow get the access and egress nodes
                nacc = iobj.getProp(ModFlow.FACC)
                negg = iobj.getProp(ModFlow.FEGG)
                ll = self.getLink(j,nacc)
                if ll != None:
                    # Write the access variables
                    val = float(self._links[ll].getProp(ModLink.LCOST))*float(iobj.getProp(ModFlow.DEM))
                    lpf.write(" + " + str(val) + " x" + str(self._xmap[(i,j)]))

                self._addLPNewLine(lpf)

                ll = self.getLink(j,negg)
                if ll != None:
                    # Write the egress variables
                    val = float(self._links[ll].getProp(ModLink.LCOST))*float(iobj.getProp(ModFlow.DEM))
                    lpf.write(" + " + str(val) + " x" + str(self._wmap[(j,i)]))

                self._addLPNewLine(lpf)

        # Write the cost of security zones: sum_js c_js^Z*y_js
        for j,jobj in self._cnodes.iteritems():
            for s in range(1, self._maxsz+1):
                val = jobj.getProp(ModNode.SZCOST+str(s))
                lpf.write(" + " + str(val) + " x" + str(self._ymap[(j,s)]))

        # Write the cost of conduits: sum_jl c_jl^C*f_jl
        for j,jobj in self._cnodes.iteritems():
            for l,lobj in self._cnodes.iteritems():

                if j == l:
                    continue

                ll = self.getLink(j,l)
                if ll == None:
                    continue

                self._addLPNewLine(lpf)

                lpf.write(" + " + self._links[ll].getProp(ModLink.CONCOST) + " x" + str(self._fmap[(j,l)]))

        lpf.write("\n")


    def _writeConstraints(self, lpf):
        '''
        Write additional constraints. Make sure we call superclass's version.
        '''

        super(ModSecDesignModel, self)._writeConstraints(lpf)

        # Write additional constraints
        self._writeAccessEgressConstraints(lpf)
        self._writeAccessEgressConstraints1(lpf)
        self._writeSecurityZoneConstraints(lpf)
        self._writeSecurityConduitConstraints(lpf)
        self._writeRealTimeConstraints(lpf)


    def _writeAccessEgressConstraints(self, lpf):
        '''
        Write access constraints: sum_j x_ij = 1, sum_j w_ji = 1
        '''

        self._cctr += 1
        strAcc = " c" + str(self._cctr) + ": "
        self._cctr += 1
        strEgg = " c" + str(self._cctr) + ": "

        first1 = True
        first2 = True
        for i,iobj in self._flows.iteritems():
            for j,jobj in self._cnodes.iteritems():

                # For each flow get the access and egress nodes
                nacc = iobj.getProp(ModFlow.FACC)
                negg = iobj.getProp(ModFlow.FEGG)
                ll = self.getLink(j,nacc)
                if ll != None:
                    # Write the access variables
                    if first1 == True:
                        strAcc += " x" + str(self._xmap[(i,j)])
                        first1 = False
                    else:
                        strAcc += " + x" + str(self._xmap[(i,j)])

                ll = self.getLink(j,negg)
                if ll != None:
                    # Write the access variables
                    if first2 == True:
                        strEgg += " x" + str(self._wmap[(j,i)])
                        first2 = False
                    else:
                        strEgg += " + x" + str(self._wmap[(j,i)])

        lpf.write(strAcc + " = 1\n")
        lpf.write(strEgg + " = 1\n")


    def _writeAccessEgressConstraints1(self, lpf):
        '''
        Write access constraints: x_ij - a_ij*g_j <= 0, w_ji - e_ji*g_j <= 0
        '''

        for i,iobj in self._flows.iteritems():
            for j,jobj in self._cnodes.iteritems():

                self._cctr += 1
                lpf.write(" c" + str(self._cctr) + ": ")
                lpf.write(" x" + str(self._xmap[(i,j)]) + " ")

                # For each flow get the access and egress nodes
                nacc = iobj.getProp(ModFlow.FACC)
                negg = iobj.getProp(ModFlow.FEGG)
                ll = self.getLink(j,nacc)
                if ll != None:
                    lpf.write(" - x" + str(self._gmap[j]))
                lpf.write(" <= 0\n")

                self._cctr += 1
                lpf.write(" c" + str(self._cctr) + ": ")
                lpf.write(" x" + str(self._wmap[(j,i)]) + " ")

                ll = self.getLink(j,negg)
                if ll != None:
                    lpf.write(" - x" + str(self._gmap[j]))
                lpf.write(" <= 0\n")


    def _writeSecurityZoneConstraints(self, lpf):
        '''
        Write security constraints: y_js - sum_i (r_is^A*x_ij + r_is^E*w_ji) <= 0
                                    alpha^Z*y_js - sum_i (r_is^A*x_ij + r_is^E*w_ji) >= 0
        '''

        for j,jobj in self._cnodes.iteritems():
            for s in range(1, self._maxsz+1):

                self._cctr += 1
                lpf.write(" c" + str(self._cctr) + ": x" + str(self._ymap[(j,s)]) + " ")

                for i,iobj in self._flows.iteritems():

                    if iobj.getProp(ModFlow.ACCSEC) == str(s):
                        lpf.write(" - x" + str(self._xmap[(i,j)]))

                    if iobj.getProp(ModFlow.EGGSEC) == str(s):
                        lpf.write(" - x" + str(self._wmap[(j,i)]))

                lpf.write(" <= 0\n")

        alpha = 100000
        for j,jobj in self._cnodes.iteritems():
            for s in range(1, self._maxsz+1):

                self._cctr += 1
                lpf.write(" c" + str(self._cctr) + ": " + str(alpha) + " x" + str(self._ymap[(j,s)]) + " ")

                for i,iobj in self._flows.iteritems():

                    if iobj.getProp(ModFlow.ACCSEC) == str(s):
                        lpf.write(" - x" + str(self._xmap[(i,j)]))

                    if iobj.getProp(ModFlow.EGGSEC) == str(s):
                        lpf.write(" - x" + str(self._wmap[(j,i)]))

                lpf.write(" >= 0\n")


    def _writeSecurityConduitConstraints(self, lpf):
        '''
        Write security constraints: f_jl - p_i*t_jl^i >= 0
                                    f_jl - alpha^C*b_jl <= 0
        '''

        for j,jobj in self._cnodes.iteritems():
            for l,lobj in self._cnodes.iteritems():

                if j == l:
                    continue

                for i,iobj in self._flows.iteritems():
                    self._cctr += 1
                    lpf.write(" c" + str(self._cctr) + ": x" + str(self._fmap[(j,l)]) + " - ")
                    if iobj.getProp(ModFlow.FSECC) == '1':
                        lpf.write(" x" + str(self._tmap[(j,l,i)]))
                    else:
                        lpf.write(iobj.getProp(ModFlow.FSECC) + " x" + str(self._tmap[(j,l,i)]))
                    lpf.write(" >= 0\n")

        alpha = 20000
        for j,jobj in self._cnodes.iteritems():
            for l,lobj in self._cnodes.iteritems():

                if j == l:
                    continue

                for i,iobj in self._flows.iteritems():
                    self._cctr += 1
                    lpf.write(" c" + str(self._cctr) + ": x" + str(self._fmap[(j,l)]))

                    if self.getLink(j,l) is not None:
                        lpf.write(" <= " + str(alpha))
                    else:
                        lpf.write(" <= 0")
                    lpf.write("\n")


    def _writeRealTimeConstraints(self, lpf):
        '''
        Write real-time constraints: sum_j (q_ij^A*x_ij + q_ji^E*w_ji) + sum_jl t_jl^i*qjl^L - q_i^M <= 0
        '''

        for i,iobj in self._flows.iteritems():

            self._cctr += 1
            lpf.write(" c" + str(self._cctr) + ": ")

            first = True
            # Write the first sum
            for j,jobj in self._cnodes.iteritems():

                # Get the nodes associated to this flow's access and egress, as well as the links
                nacc = iobj.getProp(ModFlow.FACC)
                negg = iobj.getProp(ModFlow.FEGG)

                llacc = self.getLink(nacc, j)
                llegg = self.getLink(negg, j)

                if llacc != None:
                    if first == True:
                        lpf.write(self._links[llacc].getProp(ModLink.LLAT) + " x" + str(self._xmap[(i,j)]))
                        first = False
                    else:
                        lpf.write(" + " + self._links[llacc].getProp(ModLink.LLAT) + " x" + str(self._xmap[(i,j)]))
                        
                if llegg != None:
                    if first == True:
                        lpf.write(self._links[llegg].getProp(ModLink.LLAT) + " x" + str(self._wmap[(j,i)]))
                        first = False
                    else:
                        lpf.write(" + " + self._links[llegg].getProp(ModLink.LLAT) + " x" + str(self._wmap[(j,i)]))

            # Write the second sum
            for j,jobj in self._cnodes.iteritems():
                for l,lobj in self._cnodes.iteritems():

                    if j == l:
                        continue

                    ll = self.getLink(j,l)
                    if ll == None:
                        continue

                    if first == True:
                        lpf.write(self._links[ll].getProp(ModLink.LLAT) + " x" + str(self._tmap[(j,l,i)]))
                        first = False
                    else:
                        lpf.write(" + " + self._links[ll].getProp(ModLink.LLAT) + " x" + str(self._tmap[(j,l,i)]))

            # Write the last part
            lpf.write(" <= " + self._flows[i].getProp(ModFlow.FLAT) + "\n")

