# -*- coding: utf-8 -*-
'''
NetworkDesign/mod_element/ModNSParser

@author: Bela Genge
'''
import re

class ModNSParser(object):
    '''
    '''
    NTYPE   = 'NT'   # CoS/regular
    NCOST   = 'NC'   # Node cost
    SZCOST  = 'SZC'  # Security zone cost (the number of the security zone is automatically appended)
    LN1     = 'LN1'  # Node 1 of link
    LN2     = 'LN2'  # Node 2 of link
    LCAP    = 'LCAP' # Link capacity
    LLAT    = 'LLAT' # Link latency

    def __init__(self, name):
        '''
        Constructor.
        '''
        super(ModNSParser, self).__init__()
        self.__cname = self.__class__.__name__ + ": "

        self.__nsfile = name
        self.__nodes = {}
        self.__links = {}
        self.__maxsz = -1 # The maximum number of security zones, as found in the NS file
        self.__maxsc = -1 # The maximum number of security conduits, as found in the NS file


    def parseNSFile(self):
        '''
        Parse the NS network file.
        '''

        with open(self.__nsfile, 'r') as f:
            for line in f:
                # Skip empty lines
                if len(line) <= 0:
                    continue

                # Skip comments
                if line[0] == '#':
                    continue

                if re.search('ns node', line):
                    self._parseNode(line)
                elif re.search('tb-set-node-os', line):
                    self._parseNodeOS(line)
                elif re.search('net-design-node-cost', line):
                    self._parseNodeCost(line)
                elif re.search('net-design-seczone', line):
                    self._parseSecZoneCost(line)
                elif re.search('duplex-link', line):
                    self._parseLink(line)
                elif re.search('net-design-link-conduit', line):
                    self._parseLinkConduit(line)


    def _parseNode(self, line):
        '''
        Extract node ID.
        '''

        m = re.search('set (.+) \[', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Invalid NS node declaration line: ' + line)

        nid = m.group(1)
        if nid in self.__nodes:
            raise ValueError(self.__cname + 'Node with id already exists: ' + nid)

        self.__nodes[nid] = {}

        print self.__cname + "added node: " + nid
            

    def _parseNodeOS(self, line):
        '''
        Extract the node OS - this is an important step to distinguish between CoS and other types of nodes.
        '''

        # Extract the node ID
        m = re.search(' \$(.+) ', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Invalid NS node-OS declaration line: ' + line)

        nid = m.group(1)
        if nid not in self.__nodes:
            raise ValueError(self.__cname + 'Node with id has not been declared: ' + nid)

        # Extract the node OS
        m = re.search('\$'+nid+' (.+)\n', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Cannot find NS node-OS in line: ' + line)

        nos = m.group(1)

        if nos == 'ncSCADA-CoS':
            self.__nodes[nid][ModNSParser.NTYPE] = 'CoS'
        else:
            self.__nodes[nid][ModNSParser.NTYPE] = 'N'

        print self.__cname + "added node: " + nid + ", of type: " + self.__nodes[nid][ModNSParser.NTYPE]


    def _parseNodeCost(self, line):
        '''
        Extract the node cost.
        '''

        # Extract the node ID
        m = re.search(' \$(.+) ', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Invalid NS node-cost declaration line: ' + line)

        nid = m.group(1)
        if nid not in self.__nodes:
            raise ValueError(self.__cname + 'Node with id has not been declared: ' + nid)

        # Extract the node cost
        m = re.search('\$'+nid+' (.+)\n', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Cannot find NS node-cost in line: ' + line)

        nc = m.group(1)

        # Make sure that only nodes of type CoS have costs associated
        if self.__nodes[nid][ModNSParser.NTYPE] != 'CoS':
            raise ValueError(self.__cname + 'Node that is not of type CoS cannot have a cost associated: ' + line)
    
        self.__nodes[nid][ModNSParser.NCOST] = nc

        print self.__cname + "added node: " + nid + ", of type: " + self.__nodes[nid][ModNSParser.NTYPE] + ", with cost: " + self.__nodes[nid][ModNSParser.NCOST]


    def _parseSecZoneCost(self, line):
        '''
        Extract the node cost.
        '''

        # Extract the node ID
        m = re.search(' \$(.+) ', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Invalid NS node-sz-cost declaration line: ' + line)

        nid = m.group(1)
        if nid not in self.__nodes:
            raise ValueError(self.__cname + 'Node with id has not been declared: ' + nid)

        # Extract the security zone cost
        m = re.search('\$'+nid+' (.+)\n', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Cannot find NS node-sz-cost in line: ' + line)

        szc = m.group(1)

        # Make sure that only nodes of type CoS have security zone costs associated
        if self.__nodes[nid][ModNSParser.NTYPE] != 'CoS':
            raise ValueError(self.__cname + 'Node that is not of type CoS cannot have a security zone cost associated: ' + line)
    
        # Extract the number of the security zone (we assume it is correct)
        m = re.search('seczone(\d+)\-cost', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Cannot find NS node-sz number in line: ' + line)

        szn = m.group(1)
        self.__nodes[nid][ModNSParser.SZCOST+str(szn)] = szc

        # Update the value of the maximum security zone number
        if szn > self.__maxsz:
            self.__maxsz = szn

        print self.__cname + "added node: " + nid + ", of type: " + self.__nodes[nid][ModNSParser.NTYPE] + ", with cost: " + \
                    self.__nodes[nid][ModNSParser.NCOST] + ", with sz: " +  self.__nodes[nid][ModNSParser.SZCOST+str(szn)]


    def _parseLink(self, line):
        '''
        Extract the link.
        '''

        # Split the string according to whitespace
        str = line.split()

        # Store the link ID
        lid = str[1]

        # Store the first and the second node IDs
        nid1 = str[4][1:]
        nid2 = str[5][1:]

        if (nid1 not in self.__nodes) or (nid2 not in self.__nodes):
            raise ValueError(self.__cname + 'Node(s) with id has(have) not been declared previously in line: ' + line + " (" + nid1 + "," + nid2 + ")")

        # Extract the link capacity
        m = re.search('(\d+\.\d+).+', str[6])
        if m == None:
            raise SyntaxError(self.__cname + 'Cannot find the link capacity in line: ' + line)
        lc = m.group(1)

        # Extract the link patency
        m = re.search('(\d+\.\d+).+', str[7])
        if m == None:
            raise SyntaxError(self.__cname + 'Cannot find the link latency in line: ' + line)
        ll = m.group(1)
        

        self.__links[lid] = {}
        self.__links[lid][ModNSParser.LN1] = nid1
        self.__links[lid][ModNSParser.LN2] = nid2
        self.__links[lid][ModNSParser.LCAP] = lc
        self.__links[lid][ModNSParser.LLAT] = ll

        print self.__cname + "added link: " + lid + ", N1: " + self.__links[lid][ModNSParser.LN1] + ", N2: " + self.__links[lid][ModNSParser.LN2] + \
                ", CAP: " + self.__links[lid][ModNSParser.LCAP] + ", LAT: " + self.__links[lid][ModNSParser.LLAT]


    def _parseSecConduitCost(self, line):
        '''
        Extract the cost of security conduits.
        '''

        # Extract the link ID
        m = re.search(' \$(.+) ', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Invalid NS link name declaration line: ' + line)

        lid = m.group(1)
        if lid not in self.__nodes:
            raise ValueError(self.__cname + 'The link with id has not been declared: ' + lid)

        # Extract the security conduit cost
        m = re.search('\$'+lid+' (.+)\n', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Cannot find NS link-cond-cost in line: ' + line)

        lcc = m.group(1)

        # Extract the number of the security conduit (we assume it is correct)
        m = re.search('seczone(\d+)\-cost', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Cannot find NS node-sz number in line: ' + line)

        szn = m.group(1)
        self.__nodes[nid][ModNSParser.SZCOST+str(szn)] = szc

        # Update the value of the maximum security zone number
        if szn > self.__maxsz:
            self.__maxsz = szn

        print self.__cname + "added node: " + nid + ", of type: " + self.__nodes[nid][ModNSParser.NTYPE] + ", with cost: " + \
                    self.__nodes[nid][ModNSParser.NCOST] + ", with sz: " +  self.__nodes[nid][ModNSParser.SZCOST+str(szn)]
