# -*- coding: utf-8 -*-
'''
NetworkDesign/mod_network/ModNetwork

@author: Bela Genge
'''
from mod_node import ModNode
from mod_link import ModLink
from mod_flow import ModFlow
import re

class ModNetwork(object):
    '''
    Builds the internal network configuration from an input NS file.
    '''

    def __init__(self, name):
        '''
        Constructor.
        '''
        super(ModNetwork, self).__init__()
        self.__cname = self.__class__.__name__ + ": "

        self.__nsfile = name
        self._nodes = {}
        self._cnodes = {} # Concentrator nodes
        self._links = {}
        self._flows = {}
        self._maxsz = -1 # The maximum number of security zones, as found in the NS file
        self._maxsc = -1 # The maximum level of security conduits, as found in the NS file


    def parseNSFile(self):
        '''
        Parse the NS network file.
        '''

        with open(self.__nsfile, 'r') as f:
            for line in f:
                # Skip empty lines
                if len(line) <= 0:
                    continue

                # Skip comments
                if line[0] == '#':
                    continue

                if re.search('ns node', line):
                    self._parseNode(line)
                elif re.search('tb-set-node-os', line):
                    self._parseNodeOS(line)
                elif re.search('net-design-node-cost', line):
                    self._parseNodeCost(line)
                elif re.search('net-design-seczone', line):
                    self._parseSecZoneCost(line)
                elif re.search('duplex-link', line):
                    self._parseLink(line)
                elif re.search('link-bandwidth-cost', line):
                    self._parseLinkBandwidthCost(line)
                elif re.search('net-design-link-conduit', line):
                    self._parseLinkConduitCost(line)
                elif re.search('Application/Traffic', line):
                    self._parseFlow(line)
                elif re.search('packetSize_', line):
                    self._parseFlowPacketSize(line)
                elif re.search('rate_', line):
                    self._parseFlowPacketRate(line)
                elif re.search('toleratedLatency_', line):
                    self._parseFlowToleratedLatency(line)
                elif re.search('new Agent/', line):
                    self._parseFlowSrcSink(line)
                elif re.search('\$ns attach-agent', line):
                    self._parseFlowNodes(line)
                elif re.search('security-conduit', line):
                    self._parseFlowSecurityConduit(line)
                elif re.search('access-security-zone', line):
                    self._parseAccessEgressSecZone(line, True)
                elif re.search('egress-security-zone', line):
                    self._parseAccessEgressSecZone(line, False)

        # Build the set of CoS nodes
        for j,jobj in self._nodes.iteritems():
             if jobj.getProp(ModNode.NTYPE) == ModNode.NCOS:
                self._cnodes[j] = jobj

        # Calculate demand values
        #for f,fobj in self._flows.iteritems():
        #    fobj.addProp(ModFlow.DEM, str(int(fobj.getProp(ModFlow.FPSZ))*int(fobj.getProp(ModFlow.FPRT))))


    def _parseNode(self, line):
        '''
        Extract node ID.
        '''

        m = re.search('set (.+) \[', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Invalid NS node declaration line: ' + line)

        nid = m.group(1)
        if nid in self._nodes:
            raise ValueError(self.__cname + 'Node with id already exists: ' + nid)

        self._nodes[nid] = ModNode(nid)

        print self.__cname + "added node: " + nid
            

    def _parseNodeOS(self, line):
        '''
        Extract the node OS - this is an important step to distinguish between CoS and other types of nodes.
        '''

        # Extract the node ID
        m = re.search(' \$(.+) ', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Invalid NS node-OS declaration line: ' + line)

        nid = m.group(1)
        if nid not in self._nodes:
            raise ValueError(self.__cname + 'Node with id has not been declared: ' + nid)

        # Extract the node OS
        m = re.search('\$'+nid+' (.+)\n', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Cannot find NS node-OS in line: ' + line)

        nos = m.group(1)

        if nos == ModNode.NSOSCOS:
            self._nodes[nid].addProp(ModNode.NTYPE, ModNode.NCOS)
        else:
            self._nodes[nid].addProp(ModNode.NTYPE, ModNode.NREG)

        print self.__cname + "added node: " + nid + ", of type: " + self._nodes[nid].getProp(ModNode.NTYPE)


    def _parseNodeCost(self, line):
        '''
        Extract the node cost.
        '''

        # Extract the node ID
        m = re.search(' \$(.+) ', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Invalid NS node-cost declaration line: ' + line)

        nid = m.group(1)
        if nid not in self._nodes:
            raise ValueError(self.__cname + 'Node with id has not been declared: ' + nid)

        # Extract the node cost
        m = re.search('\$'+nid+' (.+)\n', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Cannot find NS node-cost in line: ' + line)

        nc = m.group(1)

        # Make sure that only nodes of type CoS have costs associated
        if self._nodes[nid].getProp(ModNode.NTYPE) != ModNode.NCOS:
            raise ValueError(self.__cname + 'Node that is not of type CoS cannot have a cost associated: ' + line)
    
        self._nodes[nid].addProp(ModNode.NCOST, nc)

        print self.__cname + "added node: " + nid + ", of type: " + self._nodes[nid].getProp(ModNode.NTYPE) + ", with cost: " + self._nodes[nid].getProp(ModNode.NCOST)


    def _parseSecZoneCost(self, line):
        '''
        Extract the node cost.
        '''

        # Extract the node ID
        m = re.search(' \$(.+) ', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Invalid NS node-sz-cost declaration line: ' + line)

        nid = m.group(1)
        if nid not in self._nodes:
            raise ValueError(self.__cname + 'Node with id has not been declared: ' + nid)

        # Extract the security zone cost
        m = re.search('\$'+nid+' (.+)\n', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Cannot find NS node-sz-cost in line: ' + line)

        szc = m.group(1)

        # Make sure that only nodes of type CoS have security zone costs associated
        if self._nodes[nid].getProp(ModNode.NTYPE) != ModNode.NCOS:
            raise ValueError(self.__cname + 'Node that is not of type CoS cannot have a security zone cost associated: ' + line)
    
        # Extract the number of the security zone (we assume it is correct)
        m = re.search('seczone(\d+)\-cost', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Cannot find NS node-sz number in line: ' + line)

        szn = m.group(1)
        self._nodes[nid].addProp(ModNode.SZCOST+str(szn), szc)

        # Update the value of the maximum security zone number
        if int(szn) > int(self._maxsz):
            self._maxsz = int(szn)

        print self.__cname + "added node: " + nid + ", of type: " + self._nodes[nid].getProp(ModNode.NTYPE) + ", with cost: " + \
                    self._nodes[nid].getProp(ModNode.NCOST) + ", with sz: " +  self._nodes[nid].getProp(ModNode.SZCOST+str(szn))


    def _parseLink(self, line):
        '''
        Extract the link.
        '''

        # Split the string according to whitespace
        sstr = line.split()

        # Store the link ID
        lid = sstr[1]

        # Store the first and the second node IDs
        nid1 = sstr[4][1:]
        nid2 = sstr[5][1:]

        if (nid1 not in self._nodes) or (nid2 not in self._nodes):
            raise ValueError(self.__cname + 'Node(s) with id has(have) not been declared previously in line: ' + line + " (" + nid1 + "," + nid2 + ")")

        # Extract the link capacity
        m = re.search('(\d+\.\d+).+', sstr[6])
        if m == None:
            raise SyntaxError(self.__cname + 'Cannot find the link capacity in line: ' + line)
        lc = m.group(1)

        # Extract the link patency
        m = re.search('(\d+\.\d+).+', sstr[7])
        if m == None:
            raise SyntaxError(self.__cname + 'Cannot find the link latency in line: ' + line)
        ll = m.group(1)
        

        self._links[lid] = ModLink(lid)
        self._links[lid].addProp(ModLink.LN1, nid1)
        self._links[lid].addProp(ModLink.LN2, nid2)
        self._links[lid].addProp(ModLink.LCAP, lc)
        self._links[lid].addProp(ModLink.LLAT, ll)

        print self.__cname + "added link: " + lid + ", N1: " + self._links[lid].getProp(ModLink.LN1) + ", N2: " + self._links[lid].getProp(ModLink.LN2) + \
                ", CAP: " + self._links[lid].getProp(ModLink.LCAP) + ", LAT: " + self._links[lid].getProp(ModLink.LLAT)


    def _parseLinkBandwidthCost(self, line):
        '''
        Extract the unitary cost of bandwidth.
        '''

        # Extract the link ID
        m = re.search(' \$(.+) ', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Invalid NS link-cost declaration line: ' + line)

        lid = m.group(1)
        if lid not in self._links:
            raise ValueError(self.__cname + 'Node with id has not been declared: ' + lid)

        # Extract the link cost
        m = re.search('\$'+lid+' (.+)\n', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Cannot find NS link-cost in line: ' + line)

        lc = m.group(1)

        self._links[lid].addProp(ModLink.LCOST, lc)

        print self.__cname + "added link: " + lid + ", with bandwidth cost: " + self._links[lid].getProp(ModLink.LCOST)


    def _parseLinkConduitCost(self, line):
        '''
        Extract the cost of security conduits.
        '''

        # Extract the link ID
        m = re.search(' \$(.+) ', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Invalid NS link name declaration line: ' + line)

        lid = m.group(1)
        if lid not in self._links:
            raise ValueError(self.__cname + 'The link with id has not been declared: ' + lid)

        # Extract the security conduit cost
        m = re.search('\$'+lid+' (.+)\n', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Cannot find NS link-cond-cost in line: ' + line)

        lcc = m.group(1)
        self._links[lid].addProp(ModLink.CONCOST, lcc)

        print self.__cname + "added link: " + lid + ", with conduit cost: " + self._links[lid].getProp(ModLink.CONCOST)


    def _parseFlow(self, line):
        '''
        Extract the flow id.
        '''

        # Extract the flow ID
        m = re.search('set (.+) \[', line)
        if m == None:
            raise SyntaxError(self.__cname + 'Invalid NS flow name declaration line: ' + line)

        fid = m.group(1)
        if fid in self._flows:
            raise ValueError(self.__cname + 'The flow with id was declared before: ' + fid)

        self._flows[fid] = ModFlow(fid)

        print self.__cname + "added flow: " + fid 


    def _parseFlowPacketSize(self, line):
        '''
        Extract the flow's packet size (used to compute the demand value).
        '''

        # Split the line so that we can easily extract stuff
        sstr = line.split()

        # Extract the fid
        fid = sstr[0][1:]
        if fid not in self._flows:
            raise ValueError(self.__cname + 'The flow with id was not declared: ' + fid)

        # Extract the packet size (bits)
        self._flows[fid].addProp(ModFlow.FPSZ, sstr[3])

        print self.__cname + "added flow: " + fid + ", with packet size: " + sstr[3]


    def _parseFlowPacketRate(self, line):
        '''
        Extract the flow's packet rate (the demand value).
        '''

        # Split the line so that we can easily extract stuff
        sstr = line.split()

        # Extract the fid
        fid = sstr[0][1:]
        if fid not in self._flows:
            raise ValueError(self.__cname + 'The flow with id was not declared: ' + fid)

        # Extract the packet rate
        self._flows[fid].addProp(ModFlow.DEM, str(int(sstr[3])/1000))

        print self.__cname + "added flow: " + fid + ", with packet rate (demand, in Kb): " + str(int(sstr[3])/1000)


    def _parseFlowToleratedLatency(self, line):
        '''
        Extract the flow's maximum tolerated latency.
        '''

        # Split the line so that we can easily extract stuff
        sstr = line.split()

        # Extract the fid
        fid = sstr[0][1:]
        if fid not in self._flows:
            raise ValueError(self.__cname + 'The flow with id was not declared: ' + fid)

        # Extract the packet rate
        self._flows[fid].addProp(ModFlow.FLAT, sstr[3])

        print self.__cname + "added flow: " + fid + ", with maximum tolerated latency (ms): " + sstr[3]


    def _parseFlowSrcSink(self, line):
        '''
        Extract the flow's source/sink ID.
        '''

        # Split the line so that we can easily extract stuff
        sstr = line.split()

        # We distinguish between sink/src by keywords
        if re.search('src', sstr[1]):
            # We have found a source id. Now we need to extract the the fid.
            idx = sstr[1].find('src')
            fid = sstr[1][:idx]

            print self.__cname + "Found src flow id: " + sstr[1]
            self._flows[fid].addProp(ModFlow.FSRCID, sstr[1])

        elif re.search('sink', sstr[1]):
            # We have found a sink id. Now we need to extract the the fid.
            idx = sstr[1].find('sink')
            fid = sstr[1][:idx]

            print self.__cname + "Found sink flow id: " + sstr[1]
            self._flows[fid].addProp(ModFlow.FSINKID, sstr[1])

        else:
            raise ValueError(self.__cname + 'Unexpected src/sink value: ' + sstr[1])


    def _parseFlowNodes(self, line):
        '''
        Extract the access/egress nodes attached to the flow.
        '''

        # Split the line so that we can easily extract stuff
        sstr = line.split()

        # Get the node ID
        nid = sstr[2][1:]
        if nid not in self._nodes:
            raise ValueError(self.__cname + 'Node with id has not been declared: ' + nid)

        # Get the src/sink ID
        sid = sstr[3][1:]

        # Get the flow ID and the src/sink
        if re.search('src', sstr[3]):
            # We have found a source id. Now we need to extract the the fid.
            idx = sstr[3].find('src')
            fid = sstr[3][1:idx]
            self._flows[fid].addProp(ModFlow.FACC, nid)

            print self.__cname + "Access node was attached to flow: " + nid + ", " + fid

        elif re.search('sink', sstr[3]):
            # We have found a sink id. Now we need to extract the the fid.
            idx = sstr[3].find('sink')
            fid = sstr[3][1:idx]
            self._flows[fid].addProp(ModFlow.FEGG, nid)

            print self.__cname + "Egress node was attached to flow: " + nid + ", " + fid

        else:
            raise ValueError(self.__cname + 'Unexpected src/sink value: ' + sstr[3])


    def _parseFlowSecurityConduit(self, line):
        '''
        Extract the flow's source/sink ID.
        '''

        # Split the line so that we can easily extract stuff
        sstr = line.split()

        # Get the flow ID
        fid = sstr[1]
        if fid not in self._flows:
            raise ValueError(self.__cname + 'Flow with id has not been declared: ' + fid)

        # Store the security conduit
        m = re.search('Conduit(\d+)', sstr[3])
        self._flows[fid].addProp(ModFlow.FSECC, m.group(1))
        print self.__cname + "Added flow security conduit: " + fid + ", " + m.group(1)

        # Compute the maximum sec conduit
        if int(m.group(1)) > self._maxsc:
            self._maxsc = int(m.group(1))


    def _parseAccessEgressSecZone(self, line, acc):
        '''
        Extract the access/egress security zone requirement.
        '''

        # Split the line so that we can easily extract stuff
        sstr = line.split()

        # Get the agent ID and the zone ID
        aid = sstr[1]
        zid = sstr[3]

        # Get the node ID for this aid
        for (fid,fobj) in self._flows.iteritems():

            found = False
            if fobj.hasProp(ModFlow.FSRCID) and fobj.getProp(ModFlow.FSRCID) == aid:
                found = True

            if fobj.hasProp(ModFlow.FSINKID) and fobj.getProp(ModFlow.FSINKID) == aid:
                found = True
                
            if found == True:
                m = re.search('Zone(\d+)', zid)
                if acc == True:
                    fobj.addProp(ModFlow.ACCSEC, m.group(1))
                    print self.__cname + "Access security zone was added: " + aid + ", " + m.group(1)
                else:
                    fobj.addProp(ModFlow.EGGSEC, m.group(1))
                    print self.__cname + "Egress security zone was added: " + aid + ", " + m.group(1)

                return

        raise ValueError(self.__cname + 'Unable to find agent with ID: ' + aid)


    def getLink(self, n1, n2):

        for ll,llobj in self._links.iteritems():

            if (llobj.getProp(ModLink.LN1)==n1 and llobj.getProp(ModLink.LN2)==n2) or \
                (llobj.getProp(ModLink.LN2)==n1 and llobj.getProp(ModLink.LN1)==n2):
                return ll

        return None
