# -*- coding: utf-8 -*-
'''
NetworkDesign/mod_solution/ModSolution

@author: Bela Genge
'''

import SolConstants as Constants

class ModSolution(object):
    '''
    Prints the solution.
    '''

    def __init__(self, fname):
        '''
        Constructor.
        '''
        super(ModSolution, self).__init__()

        self.__nodes = []
        self.__links = []
        self.__fname = fname
        self.__cname = self.__class__.__name__ + ": "


    def writeSolution(self, map):
        '''
        Write the solution to the output XML file.
        '''
        
        of = open(self.__fname, "w")

        # Write the start of the XML
        of.write(Constants.XML_START)
        of.write("\n")

        # unpack the mapping
        (route, map_nx, map_nw, map_y, map_f) = map

        # Write stuff
        self._writeCoSNodes(of, route, map_y)
        self._writeCoSLinks(of, route, map_f)
        self._writeEndPointNodes(of, map_nx, map_nw, map_y)
        self._writeEndPointLinks(of, map_nx, map_nw)

        # Write the end of the XML
        of.write("\n")
        of.write(Constants.XML_END)

        of.close()


    def _writeCoSNodes(self, of, route, map_y):
        '''
        Write the XML for CoS nodes.
        '''

        for i,n in route.iteritems():
            for (n1,n2) in n:
                if n1 not in self.__nodes:
                    # Get the full set of zones implemented at each node
                    sz = []
                    maxsz = -1
                    for j,s in map_y:
                        if j == n1:
                            sz.append(s)
                            if s > maxsz:
                                maxsz = s
                    self._writeNode(of, n1, sz, maxsz)
                    self.__nodes.append(n1)
                if n2 not in self.__nodes:
                    sz = []
                    maxsz = -1
                    for j,s in map_y:
                        if j == n2:
                            sz.append(s)
                            if s > maxsz:
                                maxsz = s
                    self._writeNode(of, n2, sz, maxsz)
                    self.__nodes.append(n2)


    def _writeEndPointNodes(self, of, map_nx, map_nw, map_y):
        '''
        Write the XML for end-point nodes.
        '''

        for (i,(j,n)) in map_nx.iteritems():
            if n not in self.__nodes:
                self._writeEndPointNode(of, n)
                self.__nodes.append(n)
            if j not in self.__nodes:
                # This is the case in which CoS nodes are not part of at least a 2-node route
                # Get the full set of zones implemented at each node
                sz = []
                maxsz = -1
                for nc,s in map_y:
                    if j == nc:
                        sz.append(s)
                        if s > maxsz:
                            maxsz = s

                self._writeNode(of, j, sz, maxsz)
                self.__nodes.append(j)

        for (i,(j,n)) in map_nw.iteritems():
            if n not in self.__nodes:
                self._writeEndPointNode(of, n)
                self.__nodes.append(n)
            if j not in self.__nodes:
                # This is the case in which CoS nodes are not part of at least a 2-node route
                # Get the full set of zones implemented at each node
                sz = []
                maxsz = -1
                for nc,s in map_y:
                    if j == nc:
                        sz.append(s)
                        if s > maxsz:
                            maxsz = s

                self._writeNode(of, j, sz, maxsz)
                self.__nodes.append(j)



    def _writeNode(self, of, node, sz, maxsz):
        '''
        Write the XML for one node.
        '''

        nxml = Constants.XML_NODE

        ll = node + ":"
        for s in sz:
            ll += str(s) + ","
        ll = ll[0:len(ll)-1]

        # Replace identifiers in the XML with values
        nxml = nxml.replace(Constants.IXML_NID, node)
        nxml = nxml.replace(Constants.IXML_NT, ll)
        nxml = nxml.replace(Constants.IXML_NH, Constants.COS_HEIGHT)
        nxml = nxml.replace(Constants.IXML_NW, Constants.COS_WIDTH)
        nxml = nxml.replace(Constants.IXML_NS, Constants.COS_SHAPE)

        if maxsz == 1:
            nxml = nxml.replace(Constants.IXML_NF, Constants.COS_FILLZ1)
        elif maxsz == 2:
            nxml = nxml.replace(Constants.IXML_NF, Constants.COS_FILLZ2)
        elif maxsz == 3:
            nxml = nxml.replace(Constants.IXML_NF, Constants.COS_FILLZ3)
        else: # We do not really have any nodes attached to CoS, so there is no need to install a SZ
            nxml = nxml.replace(Constants.IXML_NF, Constants.COS_FILL)

        of.write(nxml)
        of.write("\n")


    def _writeEndPointNode(self, of, node):
        '''
        Write the XML for one node.
        '''

        nxml = Constants.XML_NODE

        # Replace identifiers in the XML with values
        nxml = nxml.replace(Constants.IXML_NID, node)
        nxml = nxml.replace(Constants.IXML_NT, node)
        nxml = nxml.replace(Constants.IXML_NH, Constants.REG_HEIGHT)
        nxml = nxml.replace(Constants.IXML_NW, Constants.REG_WIDTH)
        nxml = nxml.replace(Constants.IXML_NS, Constants.REG_SHAPE)
        nxml = nxml.replace(Constants.IXML_NF, Constants.REG_FILL)

        of.write(nxml)
        of.write("\n")


    def _writeCoSLinks(self, of, route, map_f):
        '''
        Write the XML for CoS links.
        '''

        # We need to pre-compute the labels.
        # Link labels will show the flows that are routed.
        llab = {}
        for i,n in route.iteritems():
            for (n1,n2) in n:
                if (n1,n2) not in llab and (n2,n1) not in llab:
                    llab[(n1,n2)] = [i]
                else:
                    if (n1,n2) in llab:
                        if i not in llab[(n1,n2)]:
                            llab[(n1,n2)].append(i)
                    else:
                        if i not in llab[(n2,n1)]:
                            llab[(n2,n1)].append(i)

        for i,n in route.iteritems():
            for (n1,n2) in n:
                if (n1,n2) not in self.__links and (n2,n1) not in self.__links:
                    ll = ''
                    if (n1,n2) in llab:
                        for s in llab[(n1,n2)]:
                            ll += s + ","
                    else:
                        for s in llab[(n2,n1)]:
                            ll += s + ","
                    ll = ll[0:len(ll)-1]
                    cond = 1
                    for ((j1,j2),cc) in map_f:
                        if (j1,j2) == (n1,n2) or (j2,j1) == (n1,n2):
                            cond = cc
                            break

                    self._writeLink(of, ll, n1, n2, cond)
                    self.__links.append((n1,n2))



    def _writeEndPointLinks(self, of, map_nx, map_nw):
        '''
        Write the XML for end-point links.
        '''

        # We need to pre-compute the labels.
        # Link labels will show the flows that are routed.
        llab = {}
        for (i,(j,n)) in map_nx.iteritems():
            if (n,j) not in llab:
                llab[(n,j)] = [i]
            else:
                llab[(n,j)].append(i)

        for (i,(j,n)) in map_nw.iteritems():
            if (n,j) not in llab:
                llab[(n,j)] = [i]
            else:
                llab[(n,j)].append(i)

        for ((n,j),vi) in llab.iteritems():
            if (n,j) in self.__links:
                continue

            ll = ''
            for i in vi:
                ll += i + ","
            ll = ll[0:len(ll)-1]

            self._writeLink(of, ll, n, j, 0)
            self.__links.append((n,j))


    def _writeLink(self, of, i, n1, n2, cond):
        '''
        Write the XML for one link.
        '''

        lxml = Constants.XML_EDGE

        # Replace identifiers in the XML with values
        lxml = lxml.replace(Constants.IXML_EID, n1+n2)
        lxml = lxml.replace(Constants.IXML_ESRC, n1)
        lxml = lxml.replace(Constants.IXML_EDST, n2)
        lxml = lxml.replace(Constants.IXML_ELAB, str(i))
        lxml = lxml.replace(Constants.IXML_ETYPE, Constants.LINE_TYPE)

        if cond == 1:
            lxml = lxml.replace(Constants.IXML_ECOLOR, Constants.LINE_COLCON1)
            lxml = lxml.replace(Constants.IXML_EWIDTH, Constants.LINE_WCON1)
        elif cond == 2:
            lxml = lxml.replace(Constants.IXML_ECOLOR, Constants.LINE_COLCON2)
            lxml = lxml.replace(Constants.IXML_EWIDTH, Constants.LINE_WCON2)
        elif cond == 3:
            lxml = lxml.replace(Constants.IXML_ECOLOR, Constants.LINE_COLCON3)
            lxml = lxml.replace(Constants.IXML_EWIDTH, Constants.LINE_WCON3)
        else: # This could be used for reg. (end-point) nodes
            lxml = lxml.replace(Constants.IXML_ECOLOR, Constants.LINE_COL)
            lxml = lxml.replace(Constants.IXML_EWIDTH, Constants.LINE_W)

        of.write(lxml)
        of.write("\n")
